/*-----------------------------------------------------------------------------

NAME
      PatchData -- Patch the data fork of a file.

SYNOPSIS
      PatchFile file <patch-script

DESCRIPTION
      The file contents is changed according to the script given on
	  standard input.  The script syntax is as follows:
	  - Comments start with '#' and include the rest of the line.
	  - Whitespace, control characters, and comments are accepted everywhere
	    except inside keywords.
	  - Commands have the form <keyword> <hexadecimal digits> <semicolon>
	  - The commands are:
	      + size: check that the size of the file is equal to the argument.
		  + crc: compute a CRC on the file, which must match the argument.
		  + offset: set the current file position.
		  + change: check that the contents of the file at the current
		    position matches the argument.
		  + to: write the argument at the current file position.
		  + append: check that the current file position is the end of the
		    file, and append the argument to the file.
		  
	  - The size, crc, and offset commands ignore all but the eight last
	    digits of their argument.
      - The change, to, and append commands need an even number of hexadecimal
	    digits, not more than 2048 digits (1024 bytes).

WARNING
      This program will not revert the patched file to its initial state
	  in case of an error.  Script writers should always start their
	  scripts by a check of the size and CRC, and users should always make
	  a backup copy of the file before applying a patch.

NO COPYRIGHT
      This is public domain software written by Planar.  The CRC stuff was
	  written by Ross Williams.

COMPILATION
    Compile this file with the following commands (with MPW 3.2) :
      C PatchData.c
      Link -o PatchData -t MPST -c 'MPS ' PatchData.c.o 
           "{Libraries}Interface.o" "{Libraries}ToolLibs.o" 
           "{Libraries}Runtime.o" "{CLibraries}StdCLib.o"

-----------------------------------------------------------------------------*/

#include <CursorCtl.h>
#include <stdio.h>

#include "crc32.h"

long line = 1;

int get_char (void){
  int c;
  
  while (1){
    c = getchar ();
	if (c == '#'){
	  while (c != '\n') c = getchar ();
	  ++line;
	}else if (c > ' ' || c == EOF){
	  return c;
	}else if (c == '\n'){
	  ++line;
	}
  }
}

int get_digit (void){
  int c;

  c = get_char ();
  if (c >= '0' && c <= '9'){
    return c - '0';
  }else if (c >= 'a' && c <= 'f'){
    return c - 'a' + 10;
  }else if (c >= 'A' && c <= 'F'){
    return c - 'A' + 10;
  }else{
    return 1000 + c;
  }
}

unsigned long get_long (void){
  int c;
  unsigned long result = 0;
  
  while (1){
    c = get_digit ();
	if (c >= 16) return result;
	result = 16 * result + c;
  }
}

char buffer [1024];

int get_string (void){
  int d1, d2;
  int len = 0;
  
  while (1){
    d1 = get_digit ();
	if (d1 >= 16) return len;
	d2 = get_digit ();
	if (d2 >= 16){
	  fprintf (stderr, "Line %ld  # PatchData - odd number of digits\n", line);
	  return len;
	}
	if (len == 1024){
	  fprintf (stderr, "Line %ld  # PatchData - argument too long\n", line);
	  exit (3);
	}
	buffer [len] = d1 * 16 + d2;
	++len;
  }
}

char keybuf [20];
char *keys [] = {"append", "change", "crc", "offset", "size", "to"};
#define Eof -1
#define Append 0
#define Change 1
#define Crc 2
#define Offset 3
#define Size 4
#define To 5

int get_key (void){
  int c, i;
  int l = 0;
  
  c = get_char ();
  if (c == EOF) return Eof;
  while (1){
    keybuf [l] = c;
	++l;
	if (l >= 20){
	  fprintf (stderr, "Line %ld  # PatchData - bad keyword\n", line);
	  exit (3);
	}
	c = getchar ();
	if (c < 'a' || c > 'z'){
	  ungetc (c, stdin);
	  break;
	}
  }
  keybuf [l] = '\0';
  for (i = 0; i < sizeof (keys) / sizeof (keys [0]); i++){
    if (!strcmp (keybuf, keys [i])) return i;
  }
  fprintf (stderr, "Line %ld  # PatchData - bad keyword\n", line);
  exit (3);
}

int main (int argc, char *argv []){
  unsigned long cur_pos = -1, cur_len = -1, len;
  FILE *f;
  int keyword;
  char rdbuf [1024];
  
  if (argc != 2){
    fprintf (stderr, "# PatchData - usage: PatchData file <patchscript\n");
	exit (2);
  }
  f = fopen (argv [1], "rb+");
  if (f == NULL){
    fprintf (stderr, "# PatchData - cannot open \"%s\"", argv [1]);
	perror (" ");
	exit (3);
  }
  while (1){
    SpinCursor (1);
    keyword = get_key ();
	if (keyword == Eof){
	  exit (0);
	}else if (keyword == Append){
	  fseek (f, 0L, SEEK_END);
	  if (cur_pos != ftell (f)){
	    fprintf (stderr, "Line %ld  # PatchData - "
		                 "bad append (not at end of file)", line);
		exit (3);
	  }
	  len = get_string ();
	  fwrite (buffer, 1, len, f);
	}else if (keyword == Change){
	  fseek (f, cur_pos, SEEK_SET);
	  cur_len = get_string ();
	  len = fread (rdbuf, 1, cur_len, f);
	  if (len != cur_len || memcmp (buffer, rdbuf, cur_len)){
	    fprintf (stderr, "Line %ld  # PatchData - bad match\n", line);
		exit (3);
	  }
	}else if (keyword == Crc){
	  CRCaccu crc = CRCinit;
	  int c;
	  
	  fseek (f, 0L, SEEK_SET);
	  while (1){
	    c = getc (f);
		if (c == EOF) break;
		CRCupdate (crc, (unsigned char) c);
	  }
	  if (CRCfinal (crc) != get_long ()){
	    fprintf (stderr, "Line %ld  # PatchData - "
		                 "CRC value does not match (%lX)\n",
		         line, CRCfinal (crc));
		exit (3);
      }
	}else if (keyword == Offset){
	  cur_pos = get_long ();
	  cur_len = -1;
	}else if (keyword == Size){
	  fseek (f, 0L, SEEK_END);
	  if (ftell (f) != get_long ()){
	    fprintf (stderr, "Line %ld  # PatchData - Size does not match\n",
		         line);
	    exit (3);
	  }
	}else if (keyword == To){
	  len = get_string ();
	  if (len != cur_len){
	    fprintf (stderr, "Line %ld  # PatchData - "
		                 "Length of the new data is not the same as the old\n",
						 line);
        exit (3);
      }
	  fseek (f, cur_pos, SEEK_SET);
	  len = fwrite (buffer, 1, cur_len, f);
	  if (len != cur_len){
	    fprintf (stderr, "Line %ld  # PatchData - Problems writing data",
		         line);
		perror (" ");
		exit (3);
	  }
	}
  }
  return 0;  /* not reached */
}
